#include <unistd.h>
#include <getopt.h>
#include <openssl/sha.h>
#include <sstream>
#include <iomanip>
#include "SHA1.h"
#include "MD5ex.h"
#include "SHA256.h"
#include "SHA512ex.h"
#include "MD4ex.h"

using namespace std;

vector<unsigned char> StringToVector(unsigned char * str)
{
	vector<unsigned char> ret;
	for(unsigned int x = 0; x < strlen((char*)str); x++)
	{
		ret.push_back(str[x]);
	}
	return ret;
}

void DigestToRaw(string hash, unsigned char * raw)
{
	transform(hash.begin(), hash.end(), hash.begin(), ::tolower);
	string alpha("0123456789abcdef");
	for(unsigned int x = 0; x < (hash.length() / 2); x++)
	{
		raw[x] = (unsigned char)((alpha.find(hash.at((x * 2))) << 4));
		raw[x] |= (unsigned char)(alpha.find(hash.at((x * 2) + 1)));
	}
}

vector<unsigned char> * GenerateRandomData()
{
	vector<unsigned char> * ret = new vector<unsigned char>();
	int length = rand() % 128;
	for(int x = 0; x < length; x++)
	{
		ret->push_back((rand() % (126 - 32)) + 32);
	}
	return ret;
}

void TestExtender(Extender * sex)
{
	//First generate a signature, with randomly generated data
	vector<unsigned char> * vkey = GenerateRandomData();
	vector<unsigned char> * vmessage = GenerateRandomData();
	vector<unsigned char> * additionalData = GenerateRandomData();
	unsigned char * firstSig;
	unsigned char * secondSig;
	sex->GenerateSignature(*vkey, *vmessage, &firstSig);
	if(sex->ValidateSignature(*vkey, *vmessage, firstSig))
	{
		vector<unsigned char> * newData = sex->GenerateStretchedData(*vmessage, vkey->size(), firstSig, *additionalData, &secondSig);
		if(sex->ValidateSignature(*vkey, *newData, secondSig))
		{
			cout << "\t测试通过." << endl;
			delete vkey;
			delete vmessage;
			delete additionalData;
			delete newData;
			return;
		}
		else
		{
			cout << "\t生成的数据未能验证为正确签名." << endl;
			delete vkey;
			delete vmessage;
			delete additionalData;
			delete newData;
			return;
		}
	}
	else
	{
		cout << "\t初始签名检查失败." << endl;
		delete vkey;
		delete vmessage;
		delete additionalData;
		return;
	}
}

Extender * GetExtenderForHash(string sig)
{
	if(sig.length() == 40)
	{
		return new SHA1ex();
	}
	else if(sig.length() == 64)
	{
		return new SHA256ex();
	}
	else if(sig.length() == 32)
	{
		return new MD5ex();
	}
	else if(sig.length() == 128)
	{
		 return new SHA512ex();
	}
	return NULL;
}

void PrintHelp()
{
	cout << "HashPump [-h help] [-t test] [-s signature] [-d data] [-a additional] [-k keylength]" << endl;
	cout << "     HashPump 生成字符串以利用易受哈希长度扩展攻击的签名." << endl;
	cout << "     -h --help          显示此消息." << endl;
	cout << "     -t --test          运行测试以验证每个算法是否正常运行." << endl;
	cout << "     -s --signature     已知消息的签名." << endl;
	cout << "     -d --data          来自已知消息的数据." << endl;
	cout << "     -a --additional    您想要添加到已知消息中的信息." << endl;
	cout << "     -k --keylength     用于签署原始消息的密钥长度（以字节为单位）." << endl;
	cout << "     Version 1.0 with MD5, SHA1, SHA256 and SHA512 support." << endl;
	cout << "     <Developed by bwall(@bwallHatesTwits)>" << endl;
}

int main(int argc, char ** argv)
{
	string sig;
	string data;
	string input;
	ostringstream urlencode;
	int keylength = 0;
	string datatoadd;
	Extender * sex = NULL;
	bool run_tests = false;

	while(1)
	{
		int option_index = 0;
		static struct option long_options[] = {
			{"test", no_argument, 0, 0},
			{"signature", required_argument, 0, 0},
			{"data", required_argument, 0, 0},
			{"additional", required_argument, 0, 0},
			{"keylength", required_argument, 0, 0},
			{"help", no_argument, 0, 0},
			{0, 0, 0, 0}
		};

		int c = getopt_long(argc, argv, "ts:d:a:k:h", long_options, &option_index);
		if (c == -1)
			break;

		switch(c)
		{
		case 0:
			switch(option_index)
			{
			case 0:
				run_tests = true;
				break;
			case 1:
				sig.assign(optarg);
				sex = GetExtenderForHash(sig);
				if(sex == NULL)
				{
					cout << "不支持的签名大小." << endl;
					return 0;
				}
				break;
			case 2:
				data.assign(optarg);
				break;
			case 3:
				datatoadd.assign(optarg);
				break;
			case 4:
				keylength = atoi(optarg);
				break;
			case 5:
				PrintHelp();
				return 0;
			}
			break;
			case 't':
				run_tests = true;
				break;
			case 's':
				sig.assign(optarg);
				sex = GetExtenderForHash(sig);
				if(sex == NULL)
				{
					cout << "不支持的哈希大小." << endl;
					return 0;
				}
				break;
			case 'd':
				data.assign(optarg);
				break;
			case 'a':
				datatoadd.assign(optarg);
				break;
			case 'k':
				keylength = atoi(optarg);
				break;
			case 'h':
				PrintHelp();
				return 0;
		}
	}

	if(run_tests)
	{
		//Just a simple way to force tests
		cout << "测试 SHA1" << endl;
		TestExtender(new SHA1ex());

		cout << "测试 SHA256" << endl;
		TestExtender(new SHA256ex());

		cout << "测试 SHA512" << endl;
		TestExtender(new SHA512ex());

		cout << "测试 MD5" << endl;
		TestExtender(new MD5ex());

		cout << "测试 MD4" << endl;
		TestExtender(new MD4ex());

		cout << "测试 concluded" << endl;
		return 0;
	}

	if(sig.size() == 0)
	{
		cout << "输入签名: ";
		cin >> sig;
		sex = GetExtenderForHash(sig);
		if(sex == NULL)
		{
			cout << "不支持的哈希大小." << endl;
			return 0;
		}
	}

	if(data.size() == 0)
	{
		cout << "输入已知数据(没有可回车继续): ";
		cin.get();
		getline(cin, input);
		if (!input.empty())
		{
		    data = input;
		}
	}

	if(keylength == 0)
	{
		cout << "输入密钥或未知内容长度: ";
		cin >> keylength;
	}

	if(datatoadd.size() == 0)
	{
		cout << "输入要添加到尾部的数据: ";
		cin.get();
		getline(cin, input);
		if (!input.empty())
		{
		    datatoadd = input;
		}
	}

	vector<unsigned char> vmessage = StringToVector((unsigned char*)data.c_str());
	vector<unsigned char> vtoadd = StringToVector((unsigned char*)datatoadd.c_str());

	unsigned char firstSig[20];
	DigestToRaw(sig, firstSig);
	unsigned char * secondSig;
	vector<unsigned char> * secondMessage = sex->GenerateStretchedData(vmessage, keylength, firstSig, vtoadd, &secondSig);
	cout << "新数据签名: ";
	for(int x = 0; x < 20; x++)
	{
		printf("%02x", secondSig[x]);
	}
	cout << endl << "Bytes: ";
	for(unsigned int x = 0; x < secondMessage->size(); x++)
	{
		unsigned char c = secondMessage->at(x);
		if(c >= 32 && c <= 126)
		{
			cout << c;
			urlencode << static_cast<char>(c);
		}
		else
		{
			printf("\\x%02x", c);
			urlencode << "%" << std::hex << std::setw(2) << std::setfill('0') << (unsigned int) c;
		}
	}
	delete secondMessage;
	cout << endl << "Urlencode: " << urlencode.str() << endl;
	return 0;
}
